from fastapi_mqtt.config import MQTTConfig
from fastapi_mqtt.fastmqtt import FastMQTT

__author__ = 'Sabuhi Shukurov'

__email__ = 'sabuhi.shukurov@gmail.com'

credits = [
    'Sabuhi Shukurov',
    'Hasan Aliyev',
    'Tural Muradov',
    'vincentto13',
    'Jeremy T. Hetzel',
]

__all__ = ['FastMQTT', 'MQTTConfig']



import re
from typing import Any, Callable

from gmqtt import Client as MQTTClient

try:
    from uvicorn.config import logger

    log_info = logger
except ImportError:
    import logging

    log_info = logging.getLogger()


class MQTTHandlers:
    def __init__(self, client: MQTTClient, handlers: dict):
        self.client = client
        self.handlers = handlers
        self.user_message_handler = None
        self.user_connect_handler = None

    def on_message(self, handler: Callable) -> Callable[..., Any]:
        log_info.info('on_message handler accepted')
        self.user_message_handler = handler
        return handler

    def on_subscribe(self, handler: Callable) -> Callable[..., Any]:
        """
        Decorator method is used to obtain subscribed topics and properties.
        """
        log_info.info('on_subscribe handler accepted')
        self.client.on_subscribe = handler
        return handler

    def on_disconnect(self, handler: Callable) -> Callable[..., Any]:
        self.client.on_disconnect = handler
        return handler

    def on_connect(self, handler: Callable) -> Callable[..., Any]:
        log_info.info('on_connect handler accepted')
        self.user_connect_handler = handler
        return handler

    @property
    def get_user_message_handler(self):
        return self.user_message_handler

    @property
    def get_user_connect_handler(self):
        return self.user_connect_handler



from ssl import SSLContext
from typing import Optional, Union

from gmqtt.mqtt.constants import MQTTv50
from pydantic import BaseSettings as Settings


class MQTTConfig(Settings):
    """
    MQTTConfig is main the configuration to be passed client object.

    host: To connect MQTT broker, defaults to localhost
    port: To connect MQTT broker, defaults to 1883

    ssl: an SSL/TLS transport is created (by default a plain TCP transport is created)
        In case of an ssl.SSLContext object, context is used to create the transport.
        Alternately passing bool value, will set  a default context,
        which is calls ssl.create_default_context().

    keepalive: Maximum period in seconds between communications with the broker.
        This controls the rate at which the client will send ping messages.
        Defaults to 60 seconds.

    username: username for authentication, defaults to None
    password: password for authentication, defaults to None

    version: MQTT broker version to use, defaults to  MQTTv50.
        According to gmqtt.Client if the broker does not support 5.0 protocol version,
        responds with proper CONNACK reason code,
        the client will downgrade to 3.1 and reconnect automatically.

    reconnect_retries && reconnect_delay: Connected MQTT client always tries to reconnect,
        in case of lost connections. The number of reconnect attempts is unlimited.
        For changing this behavior give reconnect_retries and reconnect_delay values.
        For more info: # https://github.com/wialon/gmqtt#reconnects

    The last three parameters are used after the client disconnects abnormally

    will_message_topic: Topic of the payload
    will_message_payload: The payload
    will_delay_interval: Delay interval
    """

    host: str = 'localhost'
    port: int = 1883
    ssl: Union[bool, SSLContext] = False
    keepalive: int = 60
    username: Optional[str] = None
    password: Optional[str] = None
    version: int = MQTTv50

    reconnect_retries: Optional[int] = None
    reconnect_delay: Optional[int] = None

    will_message_topic: Optional[str] = None
    will_message_payload: Optional[str] = None
    will_delay_interval: Optional[str] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_mqtt']

package_data = \
{'': ['*']}

install_requires = \
['fastapi<1.0.0',
 'gmqtt>=0.6.11,<0.7.0',
 'pydantic>=1.9.0,<2.0.0',
 'uvicorn>=0.19.0,<0.20.0']

setup_kwargs = {
    'name': 'fastapi-mqtt',
    'version': '1.0.6',
    'description': 'fastapi-mqtt is extension for MQTT protocol',
    'long_description': '# fastapi-mqtt\n\nMQTT is a lightweight publish/subscribe messaging protocol designed for M2M (machine to machine) telemetry in low bandwidth environments.\nFastapi-mqtt is the client for working with MQTT.\n\nFor more information about MQTT, please refer to here: [MQTT](MQTT.md)\n\nFatapi-mqtt wraps around [gmqtt](https://github.com/wialon/gmqtt) module. Gmqtt Python async client for MQTT client implementation.\nModule has support of MQTT version 5.0 protocol\n\n[![MIT licensed](https://img.shields.io/github/license/sabuhish/fastapi-mqtt)](https://raw.githubusercontent.com/sabuhish/fastapi-mqtt/master/LICENSE)\n[![GitHub stars](https://img.shields.io/github/stars/sabuhish/fastapi-mqtt.svg)](https://github.com/sabuhish/fastapi-mqtt/stargazers)\n[![GitHub forks](https://img.shields.io/github/forks/sabuhish/fastapi-mqtt.svg)](https://github.com/sabuhish/fastapi-mqtt/network)\n[![GitHub issues](https://img.shields.io/github/issues-raw/sabuhish/fastapi-mqtt)](https://github.com/sabuhish/fastapi-mqtt/issues)\n[![Downloads](https://pepy.tech/badge/fastapi-mqtt)](https://pepy.tech/project/fastapi-mqtt)\n\n---\n\n## **Documentation**: [FastApi-MQTT](https://sabuhish.github.io/fastapi-mqtt/)\n\nThe key feature are:\n\nMQTT specification avaliable with help decarator methods using callbacks:\n\n- on_connect()\n- on_disconnect()\n- on_subscribe()\n- on_message()\n- subscribe(topic)\n\n- Base Settings available with `pydantic` class\n- Authetication to broker with credentials\n- unsubscribe certain topics and publish to certain topics\n\n### 🔨 Installation\n\n```sh\n $ pip install fastapi-mqtt\n```\n\n### 🕹 Guide\n\n```python\nfrom fastapi import FastAPI\nfrom fastapi_mqtt import FastMQTT, MQTTConfig\n\napp = FastAPI()\n\nmqtt_config = MQTTConfig()\n\nmqtt = FastMQTT(\n    config=mqtt_config\n)\n\nmqtt.init_app(app)\n\n\n\n@mqtt.on_connect()\ndef connect(client, flags, rc, properties):\n    mqtt.client.subscribe("/mqtt") #subscribing mqtt topic\n    print("Connected: ", client, flags, rc, properties)\n\n@mqtt.on_message()\nasync def message(client, topic, payload, qos, properties):\n    print("Received message: ",topic, payload.decode(), qos, properties)\n\n@mqtt.subscribe("my/mqtt/topic/#")\nasync def message_to_topic(client, topic, payload, qos, properties):\n    print("Received message to specific topic: ", topic, payload.decode(), qos, properties)\n\n@mqtt.on_disconnect()\ndef disconnect(client, packet, exc=None):\n    print("Disconnected")\n\n@mqtt.on_subscribe()\ndef subscribe(client, mid, qos, properties):\n    print("subscribed", client, mid, qos, properties)\n\n```\n\nPublish method:\n\n```python\nasync def func():\n    mqtt.publish("/mqtt", "Hello from Fastapi") #publishing mqtt topic\n\n    return {"result": True,"message":"Published" }\n\n\n```\n\nSubscribe method:\n\n```python\n\n@mqtt.on_connect()\ndef connect(client, flags, rc, properties):\n    mqtt.client.subscribe("/mqtt") #subscribing mqtt topic\n    print("Connected: ", client, flags, rc, properties)\n\n```\n\nChanging connection params\n\n```python\nmqtt_config = MQTTConfig(host = "mqtt.mosquito.org",\n    port= 1883,\n    keepalive = 60,\n    username="username",\n    password="strong_password")\n\n\nmqtt = FastMQTT(\n    config=mqtt_config)\n\n```\n\n# Contributing\n\nFell free to open issue and send pull request.\n\nThanks To [Contributors](https://github.com/sabuhish/fastapi-mqtt/graphs/contributors).\nContributions of any kind are welcome!\n\nBefore you start please read [CONTRIBUTING](https://github.com/sabuhish/fastapi-mqtt/blob/master/CONTRIBUTING.md)\n',
    'author': 'sabuhish',
    'author_email': 'sabuhi.shukurov@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sabuhish/fastapi-mqtt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
import uuid
from functools import partial
from typing import Any, Callable, Dict, Optional

from fastapi import FastAPI
from gmqtt import Client as MQTTClient
from gmqtt import Message
from gmqtt.mqtt.constants import MQTTv50

from .config import MQTTConfig
from .handlers import MQTTHandlers

try:
    from uvicorn.config import logger

    log_info = logger
except ImportError:
    import logging

    log_info = logging.getLogger()


class FastMQTT:
    """
    FastMQTT client sets connection parameters before connecting and manipulating the MQTT service.
    The object holds session information necessary to connect the MQTT broker.

    config: MQTTConfig config object

    client_id: Should be a unique identifier for connection to the MQTT broker.

    clean_session: Enables broker to establish a persistent session.
                            In a persistent session clean_session = False.
                            The broker stores all subscriptions for the client.
                            If the session is not persistent (clean_session = True).
                            The broker does not store anything for the client and \
                            purges all information from any previous persistent session.
                            The client_id  identifies the session.

    optimistic_acknowledgement:  #TODO more info needed
    """

    def __init__(
        self,
        config: MQTTConfig,
        *,
        client_id: Optional[str] = None,
        clean_session: bool = True,
        optimistic_acknowledgement: bool = True,
        **kwargs: Any,
    ):
        if not client_id:
            client_id = uuid.uuid4().hex

        self.client: MQTTClient = MQTTClient(client_id, **kwargs)
        self.config: MQTTConfig = config

        self.client._clean_session = clean_session
        self.client._username = config.username
        self.client._password = config.password
        self.client._host = config.host
        self.client._port = config.port
        self.client._keepalive = config.keepalive
        self.client._ssl = config.ssl
        self.client.optimistic_acknowledgement = optimistic_acknowledgement
        self.client._connect_properties = kwargs
        self.client.on_message = self.__on_message
        self.client.on_connect = self.__on_connect
        self.handlers: Dict[str, Any] = dict()
        self.mqtt_handlers = MQTTHandlers(self.client, self.handlers)
        log_info = logger

        if (
            self.config.will_message_topic
            and self.config.will_message_payload
            and self.config.will_delay_interval
        ):
            self.client._will_message = Message(
                self.config.will_message_topic,
                self.config.will_message_payload,
                self.config.will_delay_interval,
            )
            log_info.debug(
                f'topic -> {self.config.will_message_topic} \n payload -> {self.config.will_message_payload} \n will_delay_interval -> {self.config.will_delay_interval}'  # noqa E501
            )
            log_info.debug('WILL MESSAGE INITIALIZED')

    @staticmethod
    def match(topic, template):
        """
        Defined match topics

        topic: topic name
        template: template topic name that contains wildcards
        """
        topic = topic.split('/')
        template = template.split('/')

        if len(topic) < len(template):
            return False

        topic_idx = 0
        for _, part in enumerate(template):
            if part == '#':
                return True
            elif part == '+' or part == topic[topic_idx]:
                topic_idx += 1
                continue
            elif not part == topic[topic_idx]:
                return False

        if topic_idx == len(topic):
            return True
        return False

    async def connection(self) -> None:

        if self.client._username:
            self.client.set_auth_credentials(
                self.client._username, self.client._password
            )
            log_info.debug('user is authenticated')

        await self.__set_connetion_config()

        version = self.config.version or MQTTv50
        log_info.warning(f'Used broker version is {version}')

        await self.client.connect(
            self.client._host,
            self.client._port,
            self.client._ssl,
            self.client._keepalive,
            version,
        )
        log_info.debug('connected to broker..')

    async def __set_connetion_config(self) -> None:
        """
        The connected MQTT clients will always try to reconnect in case of lost connections.
        The number of reconnect attempts is unlimited.
        For changing this behavior, set reconnect_retries and reconnect_delay with its values.
        For more info: https://github.com/wialon/gmqtt#reconnects
        """
        if self.config.reconnect_retries:
            self.client.set_config(
                reconnect_retries=self.config.reconnect_retries
            )

        if self.config.reconnect_delay:
            self.client.set_config(reconnect_delay=self.config.reconnect_delay)

    def __on_connect(self, client, flags, rc, properties) -> None:
        """
        Generic on connecting handler, it would call user handler if defined.
        Will perform subscription for given topics.
        It cannot be done earlier, since subscription relies on connection.
        """
        if self.mqtt_handlers.get_user_connect_handler:
            self.mqtt_handlers.get_user_connect_handler(
                client, flags, rc, properties
            )

        for topic in self.handlers.keys():
            log_info.debug(f'Subscribing for {topic}')
            self.client.subscribe(topic)

    async def __on_message(self, client, topic, payload, qos, properties):
        """
        Generic on message handler, it will call user handler if defined.
        This will invoke per topic handlers that are subscribed for
        """
        gather = []
        if self.mqtt_handlers.get_user_message_handler:
            log_info.debug('Calling user_message_handler')
            gather.append(
                self.mqtt_handlers.get_user_message_handler(
                    client, topic, payload, qos, properties
                )
            )

        for topic_template in self.handlers.keys():
            if self.match(topic, topic_template):
                log_info.debug(f'Calling specific handler for topic {topic}')
                for handler in self.handlers[topic_template]:
                    gather.append(
                        handler(client, topic, payload, qos, properties)
                    )

        return await asyncio.gather(*gather)

    def publish(
        self,
        message_or_topic: str,
        payload: Any = None,
        qos: int = 0,
        retain: bool = False,
        **kwargs,
    ):
        """
        Defined to publish payload MQTT server

        message_or_topic: topic name

        payload: message payload

        qos: Quality of Assurance

        retain:
        """
        return self.client.publish(
            message_or_topic, payload=payload, qos=qos, retain=retain, **kwargs
        )

    def unsubscribe(self, topic: str, **kwargs):
        """
        Defined to unsubscribe topic

        topic: topic name
        """
        partial(self.client.unsubscribe, topic, **kwargs)
        log_info.debug('unsubscribe')
        if topic in self.handlers.keys():
            del self.handlers[topic]

        return self.client.unsubscribe(topic, **kwargs)

    def init_app(self, app: FastAPI) -> None:
        @app.on_event('startup')
        async def startup():
            await self.connection()

        @app.on_event('shutdown')
        async def shutdown():
            await self.client.disconnect()

    def subscribe(self, *topics) -> Callable[..., Any]:
        """
        Decorator method used to subscribe for specific topics.
        """

        def subscribe_handler(handler: Callable) -> Callable:
            log_info.debug(f'Subscribe for a topics: {topics}')
            for topic in topics:
                if topic not in self.handlers.keys():
                    self.handlers[topic] = []
                # TODO: Consider changing to weak_ref
                self.handlers[topic].append(handler)
            return handler

        return subscribe_handler

    def on_connect(self) -> Callable[..., Any]:
        """
        Decorator method used to handle the connection to MQTT.
        """

        def connect_handler(handler: Callable) -> Callable:
            log_info.debug('handler accepted')
            return self.mqtt_handlers.on_connect(handler)

        return connect_handler

    def on_message(self) -> Callable[..., Any]:
        """
        The decorator method is used to subscribe to messages from all topics.
        """

        def message_handler(handler: Callable) -> Callable:
            log_info.debug('on_message handler accepted')
            return self.mqtt_handlers.on_message(handler)

        return message_handler

    def on_disconnect(self) -> Callable[..., Any]:
        """
        The Decorator method used wrap disconnect callback.
        """

        def disconnect_handler(handler: Callable) -> Callable:
            log_info.debug('on_disconnect handler accepted')
            return self.mqtt_handlers.on_disconnect(handler)

        return disconnect_handler

    def on_subscribe(self) -> Callable[..., Any]:
        """
        Decorator method is used to obtain subscribed topics and properties.
        """

        def subscribe_handler(handler: Callable):
            log_info.debug('on_subscribe handler accepted')
            return self.mqtt_handlers.on_subscribe(handler)

        return subscribe_handler
